// Tony Urschler
// November 2, 2021

#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include "config.h"
#include "oss.h"
#include "queue.h"

// prototypes
void killandfree(int msgid);
void handle_sigint(int sig);
static void handle_timer(int sig);
static int setupinterrupt(void);
static int setuptimer(int time);
void logmsg(char *msg, int opt);
void logmemorymap(void);

// globals
int segment_id[3];
int* sysclock;
pagetable_t *pagetable;
pcb_t *pcb;
frametable_t frametable[256];
int frameallocation[8];  // bit array to keep track of open frames
int globalmsgid;
pid_t kiddos[18];
int kidallocation[1];   // bit array to keep track of open kids
int totalprocs = 0;
int aliveprocs = 0;
int nummemaccess = 0;
int numpagefaults = 0;
long totalaccesstime = 0;
int totalsegfaults = 0;

int main(int argc, char *argv[]) {
    signal(SIGINT, handle_sigint);
    signal(SIGCHLD, SIG_IGN);

    // setting error message
    char errmsgraw[100];
    snprintf(errmsgraw, 99, "%s: Error", argv[0]);
    char *errmsg = errmsgraw+2;

    // vars for argument parsing
    int c = 0;
    int pflag = 0;
    int vflag = 0;
    char* pvalue = NULL;
    char* usagestmt = "Usage: oss [-p n] [-v]";

    // validating arguments
    while ((c = getopt(argc, argv, "hvp:")) != -1) {
        switch(c) {
            case 'h':
                printf("%s\n", usagestmt);
                exit(EXIT_SUCCESS);
            case 'p':
                pflag = 1;
                pvalue = optarg;
                break;
            case 'v':
                vflag = 1;
                break;
            default:
                exit(EXIT_FAILURE);
        }
    }

    // setting verbose flag
    int verbose = vflag;

    // setting number of procs
    int maxalivechildren = 18;
    if (pflag == 1) {
        int tmp = atoi(pvalue);
        if (tmp < 18 && tmp > 0)
            maxalivechildren = tmp;
    }

    // setting up timer
    if (setupinterrupt() == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup handler for SIGPROF");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }
    if (setuptimer(10) == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup ITIMER_PROF interval timer");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }

    // allocating shared memory for system clock
    segment_id[0] = shmget(IPC_PRIVATE, 2*sizeof(int),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    if (segment_id[0] == -1) {
        perror(errmsg);
        exit(EXIT_FAILURE);
    }
    sysclock = (int*) shmat(segment_id[0], NULL, 0);
    if (sysclock == (void *)-1) {
        perror(errmsg);
        killandfree(0);
    }
    
    // sysclock[0] holds seconds and sysclock[1] holds nanoseconds
    sysclock[0] = 0;
    sysclock[1] = 0;

    // allocating shared memory for page table
    segment_id[1] = shmget(IPC_PRIVATE, 18*sizeof(pagetable_t),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    if (segment_id[1] == -1) {
        perror(errmsg);
        killandfree(0);
    }
    pagetable = (pagetable_t*) shmat(segment_id[1], NULL, 0);
    if (pagetable == (void *)-1) {
        perror(errmsg);
        killandfree(0);
    }

    // allocating shared memory for proc ctrl blk
    segment_id[2] = shmget(IPC_PRIVATE, 18*sizeof(pcb_t),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    if (segment_id[2] == -1) {
        perror(errmsg);
        killandfree(0);
    }
    pcb = (pcb_t*) shmat(segment_id[2], NULL, 0);
    if (pcb == (void *)-1) {
        perror(errmsg);
        killandfree(0);
    }

    // initializing message queue
    key_t key;
    key = ftok("testprog", 65);
    globalmsgid = msgget(key, 0666 | IPC_CREAT);
    if (globalmsgid == -1) {
        perror(errmsg);
        killandfree(0);
    }
    msg_t msg;
    int msgid = globalmsgid;

    // sterilize frame table
    frameallocation[0] = 0;
    for (int i = 0; i < 256; i++) {
        frametable[i].entry = 0;
    }

    // making ids into strings for children
    char pagetableBuff[50], clockBuff[50], msgIdBuff[50];
    snprintf(pagetableBuff, 49, "%d", segment_id[1]);
    snprintf(clockBuff, 49, "%d", segment_id[0]);
    char *clockstr = clockBuff;
    char *pagetablestr = pagetableBuff;

    // buffer and pointer for logging messages
    char logbuffer[400];
    char *message;

    // seeding rand
    srand(getpid());

    // times for another child to be made
    int makebabysec = 0;
    int makebabyns = 0;

    // time to shift bits
    int nextshiftns = 100000000;
    int nextshiftsec = 0;

    // total number of valid frames
    int totalallocated = 0;

    // setting up queue
    struct Queue* queue = createQueue(1000);
    int sizeofqueue = 0;

    // log variables
    logmsg("Starting simulation...", 1);
    int lines = 1;
    int logmapsec = 1;
    
    // main loop    
    for (;;) {
        // creating new child
        if (aliveprocs == 0 || (totalprocs < 42 &&
                ((makebabysec == sysclock[0] && makebabyns <= sysclock[1]) 
                 || (makebabysec < sysclock[0])))) {
            int i;
            // find a place to store child info
            for (i = 0; i < maxalivechildren; i++)
                if (TestBit(kidallocation, i) == 0)
                    break;

            // setup pagetable, pcb, and create child
            if (i < maxalivechildren) {
                SetBit(kidallocation, i);
                char lpidbuff[5];
                snprintf(lpidbuff, 4, "%d", i);
                char *lpidstr = lpidbuff;
                int size = (rand() % (32 - 1)) + 1;
                pagetable[i].delimeter = size;
                snprintf(logbuffer, 299,
                        "Master: creating P%d at time %d:%d",
                        i, sysclock[0], sysclock[1]);
                message = logbuffer;
                logmsg(message, 0);
                snprintf(msgIdBuff, 49, "%d", msgid);
                char *msgidstr = msgIdBuff;
                kiddos[i] = fork();
                if (kiddos[i] == 0)
                    execlp("./user", "./user", pagetablestr, clockstr,
                            msgidstr, lpidstr, NULL);
                totalprocs++;
                aliveprocs++;
                
                // set time for next child to be made
                makebabyns += (rand() % 499000000) + 1000000 + sysclock[0];
                makebabysec = sysclock[0];
                if (makebabyns >= 1000000000) {
                    makebabyns -= 1000000000;
                    makebabysec += 1;
                }
                sysclock[1] += 100000;
                if (sysclock[1] >= 1000000000) {
                    sysclock[1] -= 1000000000;
                    sysclock[0] += 1;
                }
            }
        }

        // check queue to see if io is complete
        if (isEmpty(queue) != 1) {

            // if the first one in queue is ready to dequeue
            if (pcb[peak(queue)].waitsec < sysclock[0] ||
                    (pcb[peak(queue)].waitns <= sysclock[1] &&
                     pcb[peak(queue)].waitsec == sysclock[0])) {

                // dequeue and log info depending on r/w request
                int id = dequeue(queue);
                if (pcb[id].requesttype == 'R') {   // if read
                    snprintf(logbuffer, 299,
                            "Master: Giving data to P%d at %d:%d", 
                            id, sysclock[0], sysclock[1]);
                    sysclock[1] += 10;
                    if (sysclock[1] > 1000000000) {
                        sysclock[1] -= 1000000000;
                        sysclock[0] += 1;
                    }
                    pcb[id].memaccesstime += 10;
                    totalaccesstime += 14000010;
                }
                // if write take more time and set dirty bit
                else {   
                    sysclock[1] += 100000;
                    if (sysclock[1] > 1000000000) {
                        sysclock[1] -= 1000000000;
                        sysclock[0] += 1;
                    }
                    pcb[id].memaccesstime += 100000;
                    totalaccesstime += 14100000;
                    snprintf(logbuffer, 299,
                        "Master: Dirty bit of frame %d set, adding time to"
                        " the clock.\nMaster: Telling P%d that write has occurred"
                        " to address %.5d", pcb[id].frame, id, pcb[id].logical);
                    SetBit(pagetable[id].dirty, pcb[id].page);
                    SetBitVar(frametable[pagetable[id].entry[pcb[id].page]].dirty, 0);
                }
                nummemaccess++;
                message = logbuffer;
                if (verbose == 1) {
                    logmsg(message, 0);
                    lines++;
                }

                // send messge
                snprintf(msg.mtext, 19, "%d", pcb[id].page);
                msg.mtype = id+2;
                int z = msgsnd(msgid, &msg, sizeof(msg), 0);
                if (z == -1) {
                    perror(errmsg);
                    killandfree(msgid);
                }
                sizeofqueue--;
            }
        }

        // check if all processes are in the queue
        if (sizeofqueue == aliveprocs) {
            int id = dequeue(queue);
            sysclock[0] = pcb[id].waitsec;
            sysclock[1] = pcb[id].waitns;
            if (pcb[id].requesttype == 'R') {
                snprintf(logbuffer, 299,
                        "Master: Waking up and giving data to P%d at %d:%d", 
                        id, sysclock[0], sysclock[1]);
                sysclock[1] += 10;
                if (sysclock[1] > 1000000000) {
                    sysclock[1] -= 1000000000;
                    sysclock[0] += 1;
                }
                pcb[id].memaccesstime += 10;
                totalaccesstime += 100010;
            }
            // if write take more time and set dirty bit
            else {
                sysclock[1] += 100000;
                if (sysclock[1] > 1000000000) {
                    sysclock[1] -= 1000000000;
                    sysclock[0] += 1;
                }
                pcb[id].memaccesstime += 100000;
                totalaccesstime += 200000;
                snprintf(logbuffer, 299,
                    "Master: Dirty bit of frame %d set, adding time to"
                    " the clock.\nMaster: Waking up and telling P%d that write"
                    " has occurred to address %.5d", 
                    pcb[id].frame, id, pcb[id].logical);
                SetBit(pagetable[id].dirty, pcb[id].page);
                SetBitVar(frametable[pagetable[id].entry[pcb[id].page]].dirty, 0);
            }
            nummemaccess++;
            message = logbuffer;
            if (verbose == 1) {
                logmsg(message, 0);
                lines++;
            }

            // send messge
            snprintf(msg.mtext, 19, "%d", pcb[id].page);
            msg.mtype = id+2;
            int z = msgsnd(msgid, &msg, sizeof(msg), 0);
            if (z == -1) {
                perror(errmsg);
                killandfree(msgid);
            }
            sizeofqueue--;
        }

        // wait for message from child
        msgrcv(msgid, &msg, sizeof(msg), 22, 0);

        // break apart message
        int lpid = atoi(strtok(msg.mtext, " "));
        char *requesttype = strtok(NULL, " ");
        int logical = atoi(strtok(NULL, " "));

        // get page number
        int page = logical >> 10;
        
        // check if process terminated or invalid request
        if (requesttype[0] == 'T' || page > pagetable[lpid].delimeter) {
            // if an invalid memory access
            if (requesttype[0] != 'T') {
                kill(kiddos[lpid], SIGINT);
                totalsegfaults++;
            }

            // log to file
            snprintf(logbuffer, 299,
                    "Master: P%d terminated. Memory access time %dns."
                    " Clearing memory at %d:%d",
                    lpid, pcb[lpid].memaccesstime, sysclock[0], sysclock[1]);
            message = logbuffer;
            logmsg(message, 0);
            lines++;

            // clear bit in child bit array
            ClearBit(kidallocation, lpid);

            // make all allocated frames available
            for (int m = 0; m < 32; m++) {
                if (TestBit(pagetable[lpid].present, m) == 1) {
                    frametable[pagetable[lpid].entry[m]].entry = 0;
                    frametable[pagetable[lpid].entry[m]].dirty = 0;
                    ClearBit(frameallocation, pagetable[lpid].entry[m]);
                    totalallocated--;
                }
                pagetable[lpid].entry[m] = 0;
            }


            // reset page table
            pagetable[lpid].dirty[0] = 0;
            pagetable[lpid].present[0] = 0;

            // increment clock
            sysclock[1] += 1000000;
            if (sysclock[1] > 1000000000) {
                sysclock[1] -= 1000000000;
                sysclock[0] += 1;
            }
            continue;
        }

        snprintf(logbuffer, 299,
                "Master: P%d requesting %s of address %.5d at %d:%d",
                lpid, requesttype, logical, sysclock[0], sysclock[1]);
        message = logbuffer;
        if (verbose == 1) {
            logmsg(message, 0);
            lines++;
        }

        // check if in memory
        if (TestBit(pagetable[lpid].present, page)) {
            sysclock[1] += 10;
            if (sysclock[1] > 1000000000) {
                sysclock[1] -= 1000000000;
                sysclock[0] += 1;
            }
            pcb[lpid].memaccesstime += 10;

            // if a read request
            if (requesttype[0] == 'R') {
                snprintf(logbuffer, 299,
                        "Master: Address %.5d in frame %d. Giving data to P%d"
                        " at %d:%d", logical, pagetable[lpid].entry[page], lpid,
                        sysclock[0], sysclock[1]);
                totalaccesstime += 10;
            }

            // if a write request, take more time
            else {
                sysclock[1] += 100000;
                if (sysclock[1] > 1000000000) {
                    sysclock[1] -= 1000000000;
                    sysclock[0] += 1;
                }
                pcb[lpid].memaccesstime += 100000;
                totalaccesstime += 100010;
                snprintf(logbuffer, 299,
                    "Master: Dirty bit of frame %d set, adding time to"
                    " the clock.\nMaster: Telling P%d that write has occurred to"
                    " address %.5d", pagetable[lpid].entry[page], lpid, logical);
                SetBit(pagetable[lpid].dirty, page);
                SetBitVar(frametable[pagetable[lpid].entry[page]].dirty, 0);
            }

            // update ref byte
            SetBitVar(frametable[pagetable[lpid].entry[page]].entry, 7);

            // set valid bit
            SetBitVar(frametable[pagetable[lpid].entry[page]].valid, 0);

            nummemaccess++;
            message = logbuffer;
            if (verbose == 1) {
                logmsg(message, 0);
                lines++;
            }

            // send message to child
            msg.mtype = lpid + 2;
            snprintf(msg.mtext, 19, "%d", page);
            int z = msgsnd(msgid, &msg, sizeof(msg), 0);
            if (z == -1) {
                perror(errmsg);
                killandfree(msgid);
            }
        }

        // if not in memory
        else {
            numpagefaults++;

            // set wait time
            pcb[lpid].waitsec = sysclock[0];
            pcb[lpid].waitns = 14000000 + sysclock[1];
            if (pcb[lpid].waitns >= 1000000000) {
                pcb[lpid].waitns -= 1000000000;
                pcb[lpid].waitsec += 1;
            }

            // store page and request type
            pcb[lpid].page = page;
            pcb[lpid].requesttype = requesttype[0];
            pcb[lpid].logical = logical;
            
            // enqueue process
            enqueue(queue, lpid);
            sizeofqueue++;

            snprintf(logbuffer, 299,
                    "Master: Address %.5d not in a frame, pagefault,"
                    " adding %d to wait queue", logical, lpid);
            message = logbuffer;
            if (verbose == 1) {
                logmsg(message, 0);
                lines++;
            }

            int j;
            // try and find a free frame
            for (j = 0; j < 256; j++)
                if (TestBit(frameallocation, j) == 0)
                    break;

            // if there's a free frame
            // set appropriate bits and
            // store info in tables
            if (j < 256) {
                pcb[lpid].frame = j;
                snprintf(logbuffer, 299,
                        "Master: Frame %d is free. Storing P%d Page %d",
                        j, lpid, page);
                message = logbuffer;
                if (verbose == 1) {
                    logmsg(message, 0);
                    lines++;
                }
                SetBit(frameallocation, j);
                SetBit(pagetable[lpid].present, page);
                pagetable[lpid].entry[page] = j;
                frametable[j].owner = lpid;
                SetBitVar(frametable[j].entry, 7);
            
                // set valid bit
                SetBitVar(frametable[j].valid, 0);
                totalallocated++;
            }
            // else see if there's an invalid bit marked by daemon
            // or run replacement algorithm
            else {
                int flag = 0;
                // check for invalid
                for (int m = 0; m < 256; m++) {
                    if (TestBitVar(frametable[m].valid, 0) == 0) {
                        // update old proc's pagetable
                        int n;
                        for (n = 0; n < 32; n++) {
                            if(pagetable[frametable[m].owner].entry[n] ==
                                    frametable[m].entry)
                                break;
                        }
                        ClearBit(pagetable[frametable[m].owner].present, n);
                        
                        // update frame table, page table pcb with new info
                        pcb[lpid].frame = m;
                        frametable[m].owner = lpid;
                        frametable[m].entry = 0;
                        SetBitVar(frametable[m].entry, 7);
                        
                        // set valid bit
                        SetBitVar(frametable[m].valid, 0);
                        
                        // log info
                        snprintf(logbuffer, 299,
                                "Master: Clearing frame %d and swapping in"
                                " P%d Page %d", m, lpid, page);
                        message = logbuffer;
                        logmsg(message, 0);
                        lines++;

                        flag = 1;
                        break;
                    }
                }

                // if needed to use replacement algorithm
                if (flag == 0) {
                    // find least recently used frame
                    int smallest = frametable[0].entry;
                    int index = 0;
                    for (int m = 0; m < 256; m++) {
                        if (frametable[m].entry > smallest) {
                            smallest = frametable[m].entry;
                            index = m;
                        }
                    }

                    // if it has a dirty bit take more time
                    if (TestBit(pagetable[frametable[index].owner].dirty, 0) != 0) {
                        sysclock[1] += (rand() % 100000) + 50000;
                        if (sysclock[1] > 1000000000) {
                            sysclock[1] -= 1000000000;
                            sysclock[0] += 1;
                        }
                    }
                    // if its clean take less time
                    else {
                        sysclock[1] += 1000;
                        if (sysclock[1] > 1000000000) {
                            sysclock[1] -= 1000000000;
                            sysclock[0] += 1;
                        }
                    }

                    // update old proc's pagetable
                    int n;
                    for (n = 0; n < 32; n++) {
                        if(pagetable[frametable[index].owner].entry[n] ==
                                frametable[index].entry)
                            break;
                    }
                    ClearBit(pagetable[frametable[index].owner].present, n);
                    
                    // update frame table, page table pcb with new info
                    pcb[lpid].frame = index;
                    frametable[index].owner = lpid;
                    frametable[index].entry = 0;
                    SetBitVar(frametable[index].entry, 7);
                    
                    // set valid bit
                    SetBitVar(frametable[index].valid, 0);
            
                    // log info
                    snprintf(logbuffer, 299,
                            "Master: Clearing frame %d and swapping in"
                            " P%d Page %d", index, lpid, page);
                    message = logbuffer;
                    logmsg(message, 0);
                    lines++;
                }
            }
        }

        // increment system clock
        sysclock[1] += (rand() % 1000000) + 500000;
        if (sysclock[1] >= 1000000000) {
            sysclock[1] -= 1000000000;
            sysclock[0] += 1;
        }

        // check how many lines are in logfile
        if (lines >= 100000) {
            printf("Max lines in log file hit.\n");
            break;
        }

        // check if its time to shift bits
        if (nextshiftns <= sysclock[1] && nextshiftsec == sysclock[0]
                || nextshiftsec < sysclock[0]) {
            // shift all bits
            for (int i = 0; i < 256; i++) {
                if (TestBit(frameallocation, i) != 0)
                    frametable[i].entry = frametable[i].entry >> 1;
            }

            // set next shift time
            nextshiftns = sysclock[1] + 100000000;
            nextshiftsec = sysclock[0];
            if (nextshiftns > 1000000000) {
                nextshiftns -= 1000000000;
                nextshiftsec += 1;
            }

        }

        // check if its time to call the daemon
        if (totalallocated > 230) {
            // find the 13 smallest elements
            int smallest[] = {
                INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX,
                INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX,
                INT_MAX, INT_MAX, INT_MAX
            };
            int indecies[13];
            for (int m = 0; m < 256; m++) {
                for (int n = 0; n < 13; n++)
                    if (frametable[m].entry < smallest[n]) {
                        smallest[n] = frametable[m].entry;
                        indecies[n] = n;
                    }
            }
            for (int m = 0; m < 13; m++) {
                ClearBitVar(frametable[indecies[m]].valid, 0);
            }
        }

        // check if its time to print mem map
        if (logmapsec <= sysclock[0]) {
            logmapsec = sysclock[0] + 1;
            logmemorymap();
        }

        // testing termination criteria
        if (totalprocs > 40)
            break;
        
    }
    sleep(.5);

    killandfree(msgid);
}

// kills all child processes and frees all
// shared memory
void killandfree(int msgid) {
    for (int i = 0; i < 18; i++) {
        if (kiddos[i] > 0)
            kill(kiddos[i], SIGINT);
    }

    // logging stats
    char statbuff[200];
    char* statmsg;
    FILE *fptr;
    fptr = fopen("logfile", "a");

    double totalseconds = sysclock[0] 
                          + (sysclock[1] / 1000000000);

    fprintf(fptr, "\n ****** STATS ******\n");

    // mem acc per second
    snprintf(statbuff, 199, "Memory accesses per second:"
            "%f", (double)nummemaccess/totalseconds);
    statmsg = statbuff;
    fprintf(fptr, "%s\n", statmsg);
    
    // page faults per mem acc
    snprintf(statbuff, 199, "Page fault per memory"
            " access: %f", 
            (double)numpagefaults/nummemaccess);
    statmsg = statbuff;
    fprintf(fptr, "%s\n", statmsg);
    
    // avg mem acc speed
    snprintf(statbuff, 199, "Average access speed"
            " (ns/access): %f",
           (double)totalaccesstime/nummemaccess);
    statmsg = statbuff;
    fprintf(fptr, "%s\n", statmsg);

    // seg fault per mem acc
    snprintf(statbuff, 199, "Seg faults per memory"
            " access: %f", 
            (double)totalsegfaults/nummemaccess);
    statmsg = statbuff;
    fprintf(fptr, "%s\n", statmsg);
    
    fclose(fptr);
    
    // freeing shared memory
    shmdt(sysclock);
    shmdt(pagetable);
    shmdt(pcb);

    shmctl(segment_id[0], IPC_RMID, 0);
    shmctl(segment_id[1], IPC_RMID, 0);
    shmctl(segment_id[2], IPC_RMID, 0);
    
    msgctl(globalmsgid, IPC_RMID, NULL);

    exit(EXIT_SUCCESS);
}

// called when SIGINT is received
void handle_sigint(int sig) {
    printf("\nSIGINT caught!\n");
    killandfree(0);
}

// used to log messages to logfile
void logmsg(char* msg, int opt) {
    if (opt != 1) {
        FILE *fptr;
        fptr = fopen("logfile", "a");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
    else {
        FILE *fptr;
        fptr = fopen("logfile", "w");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
}

// called when the timer interval has occurred
void handle_timer(int sig) {
    printf("Times up! Killing children and freeing shared memory.\n");
    killandfree(globalmsgid);
}

// sets up the interrupt for timer interval
static int setupinterrupt(void) {
    struct sigaction act;
    act.sa_handler = handle_timer;
    act.sa_flags = 0;
    return (sigemptyset(&act.sa_mask) || sigaction(SIGALRM, &act, NULL));
}

// sets up timer
static int setuptimer(int time) {
    struct itimerval value;
    value.it_interval.tv_sec = time;
    value.it_interval.tv_usec = 0;
    value.it_value = value.it_interval;
    return (setitimer(ITIMER_REAL, &value, NULL));
}

// logs the memory map
void logmemorymap(void) {
    // open file
    FILE *fptr;
    fptr = fopen("logfile", "a");

    // write to the log
    char line[100];
    char* linemsg;
    snprintf(line, 99, "\nCurrent memory layout at time %d:%d is:",
            sysclock[0], sysclock[1]);
    linemsg = line;
    fprintf(fptr, "%s\n", linemsg);

    snprintf(line, 99, "           Occupied   RefByte   DirtyBit");
    linemsg = line;
    fprintf(fptr, "%s\n", linemsg);

    char* occupied = "No";
    for (int i = 0; i < 256; i++) {
        occupied = "No";
        if (TestBit(frameallocation, i) != 0)
            occupied = "Yes";
        snprintf(line, 99, "Frame %3d:%9s%10d%11d",
                i, occupied, frametable[i].entry, frametable[i].dirty);
        linemsg = line;
        fprintf(fptr, "%s\n", linemsg);
    }
    fprintf(fptr, "\n");
    fclose(fptr);
}

